@@ -1,8 +1,10 @@
-__version__ = '0.94.1'
+# __version__ = '0.94.1'
 
-from . import download
+from .version import version as __version__
 from . import templates
 from .core import *
 from .extern import *
 from .util import *
 from . import data
+from . import db
+from . import download
