@@ -10,7 +10,7 @@ import numpy as np
 import eispac.core.fitting_functions as fit_fns
 from eispac.core.eiscube import EISCube
 from eispac.core.read_cube import read_cube
-from eispac.core.read_template import EISFitTemplate
+from eispac.core.eisfittemplate import EISFitTemplate
 from eispac.core.eisfitresult import EISFitResult
 from eispac.core.eisfitresult import create_fit_dict
 from eispac.core.scale_guess import scale_guess
@@ -85,8 +85,8 @@ def fit_with_mpfit(wave_cube, inten_cube, errs_cube, template, parinfo,
     # Extract fit information from template
     n_gauss = template['n_gauss']
     n_poly = template['n_poly']
-    min_wave = template['data_x'][0]
-    max_wave = template['data_x'][-1]
+    min_wave = template['wmin'] #template['data_x'][0]
+    max_wave = template['wmax'] #template['data_x'][-1]
     oldguess = template['fit']
 
     # Create fit dictionary, mask array, and parameter indices
@@ -167,6 +167,8 @@ def fit_with_mpfit(wave_cube, inten_cube, errs_cube, template, parinfo,
                 fit_dict['perror'][ii,jj,:] = out.perror
                 fit_dict['int'][ii,jj,:] = l_inten
                 fit_dict['err_int'][ii,jj,:] = e_inten
+                fit_dict['width'][ii,jj,:] = fwdths
+                fit_dict['err_width'][ii,jj,:] = ewdths
             else:
                 print(' ! fit did not converge!')
                 fit_dict['status'][ii,jj] = out.status
@@ -241,6 +243,8 @@ def fit_spectra(inten, template, parinfo=None, wave=None, errs=None, min_points=
         parinfo_copy = copy.deepcopy(template_obj.parinfo)
         tmplt_filename = template_obj.filename_temp
     elif isinstance(template, EISFitTemplate):
+        template._validate_parinfo_list() #force validation of .parinfo
+        template._validate_template_dict() #force check/update of .template
         template_copy = copy.deepcopy(template.template)
         parinfo_copy = copy.deepcopy(template.parinfo)
         tmplt_filename = template.filename_temp
@@ -410,6 +414,8 @@ def fit_spectra(inten, template, parinfo=None, wave=None, errs=None, min_points=
                 fit_res.fit['perror'][:,jj,:] = pool_out[jj]['perror'][:,0,:]
                 fit_res.fit['int'][:,jj,:] = pool_out[jj]['int'][:,0,:]
                 fit_res.fit['err_int'][:,jj,:] = pool_out[jj]['err_int'][:,0,:]
+                fit_res.fit['width'][:,jj,:] = pool_out[jj]['width'][:,0,:]
+                fit_res.fit['err_width'][:,jj,:] = pool_out[jj]['err_width'][:,0,:]
 
         # Calculate the Doppler velocity for each line
         # TODO: revisit error estimation
