@@ -8,12 +8,13 @@ import numpy as np
 import h5py
 import astropy.wcs
 import astropy.units as u
+from astropy.nddata import StdDevUncertainty
 import sunpy.coordinates as coords
-from eispac.core.eiscube import EISCube
+from eispac.core.eiscube import EISCube, _make_wcs_from_cube_index
 from eispac.core.read_wininfo import read_wininfo
 from eispac.instr.calc_read_noise import calc_read_noise
 
-def read_cube(filename=None, window=0, apply_radcal=True, radcal=None,
+def read_cube(filename=None, window=0, exp_set='sum', apply_radcal=True, radcal=None,
               abs_errs=True, count_offset=None, debug=False):
     """Load a single window of EIS data from an HDF5 file into an EISCube object
 
@@ -24,6 +25,11 @@ def read_cube(filename=None, window=0, apply_radcal=True, radcal=None,
     window : int, float, or str, optional
         Requested spectral window number or the value of any wavelength within
         the requested window. Default is '0'
+    exp_set : int or str, optional
+        Index number of the exposure set to load from the file. Only used in the 
+        rare case of scanning observations with multiple exposures per raster 
+        position. Can also set to a value of "sum" to add together all exposures 
+        and simulate a longer exposure image. Default is "sum".
     apply_radcal : bool, optional
         If set to True, will apply the pre-flight radiometric calibration curve
         found in the HDF5 header file and set units to erg/(cm^2 s sr). If set
@@ -105,7 +111,7 @@ def read_cube(filename=None, window=0, apply_radcal=True, radcal=None,
     if int(window) < 25:
         # Interpret values < 25 as window number
         if window >=0 and window < num_win:
-            meta['iwin'] = window
+            meta['iwin'] = int(window)
             meta['iwin_str'] = f'win{window:02d}'
             print(f'Found window {window}')
         else:
@@ -119,7 +125,7 @@ def read_cube(filename=None, window=0, apply_radcal=True, radcal=None,
         p = (wininfo['wvl_max'] - wvl)*(wvl - wininfo['wvl_min'])
         iwin = np.where(p >= 0)[0]
         if len(iwin) == 1:
-            meta['iwin'] = iwin[0]
+            meta['iwin'] = int(iwin[0])
             meta['iwin_str'] = f'win{iwin[0]:02d}'
             print(f'Found a wavelength {wvl:.2f} [Angstroms] in window {iwin[0]}')
         else:
@@ -141,9 +147,10 @@ def read_cube(filename=None, window=0, apply_radcal=True, radcal=None,
         for key in f_head['index']:
             val = np.array(f_head['index/'+key])
             if type(val[0]) == np.bytes_:
-                val = val.astype(np.unicode_) # convert bytes to unicode
+                val = val.astype(np.str_) # convert bytes to unicode
             if val.size == 1:
-                val = val[0]
+                # val = val[0]
+                val = val.item() # Extract single-value to a Python scaler
             index[key] = val
 
         meta['index'] = index
@@ -153,9 +160,10 @@ def read_cube(filename=None, window=0, apply_radcal=True, radcal=None,
         for key in f_head['pointing']:
             val = np.array(f_head['pointing/'+key])
             if type(val[0]) == np.bytes_:
-                val = val.astype(np.unicode_) # convert bytes to unicode
+                val = val.astype(np.str_) # convert bytes to unicode
             if val.size == 1:
-                val = val[0]
+                # val = val[0]
+                val = val.item() # Extract single-value to a Python scaler
             pointing[key] = val
 
         meta['pointing'] = pointing
@@ -175,11 +183,26 @@ def read_cube(filename=None, window=0, apply_radcal=True, radcal=None,
 
         # Read time and duration information
         try:
-            meta['date_obs'] = np.array(f_head['times/date_obs']).astype(np.unicode_)
-            meta['date_obs_format'] = np.array(f_head['times/time_format']).astype(np.unicode_)[0]
+            meta['date_obs'] = np.array(f_head['times/date_obs']).astype(np.str_)
+            meta['date_obs_format'] = str(np.array(f_head['times/time_format']).astype(np.str_)[0])
         except KeyError:
-            print('WARNING: complete date_obs information for each raster step'
-                 +' is missing in the HDF5 header file!')
+            # Estimate missing or broken timestamps (ideally, never used)
+            print('WARNING: the header file has missing or incomplete date_obs' 
+                 +' for each raster step! Filling with estimated timestamps.')
+            total_time = np.datetime64(index['date_end']) - np.datetime64(index['date_obs'])
+            total_time = total_time / np.timedelta64(1, 's') # ensure [s] units
+            n_xsteps = index['nexp'] / index['nexp_prp']
+            est_cad = total_time / n_xsteps
+            est_date_obs = np.datetime64(index['date_obs']) \
+                         + np.arange(n_xsteps) \
+                         * np.timedelta64(int(est_cad*1000), 'ms')
+            if index['nraster'] == 1:
+                # Sit-and-stare timestamps inc left to right
+                meta['date_obs'] = est_date_obs.astype('str').astype('<U24')
+            else:
+                # raster timestamps inc from right to left (scan dir)
+                meta['date_obs'] = np.flip(est_date_obs.astype('str').astype('<U24'))
+            meta['date_obs_format'] = 'iso_8601'
         meta['duration'] = np.array(f_head['exposure_times/duration'])
         step_duration_units = f_head['exposure_times/duration_units'][0]
         meta['duration_units'] = step_duration_units.decode('utf-8')
@@ -207,6 +230,73 @@ def read_cube(filename=None, window=0, apply_radcal=True, radcal=None,
     else:
         radcal_array = None
 
+    ############################################################################
+    ### Check for multiple exposure sets and select one (very rarely used)
+    ############################################################################
+    if len(lv_1_counts.shape) == 4:
+        nexp_per_pos = lv_1_counts.shape[0] # Num exposures per slit position
+
+        if not isinstance(exp_set, (str, int)):
+            print(f"ERROR: invalid 'raster' keyword data type. Please input a"
+                 +f" string of 'sum' or an integer.", file=sys.stderr)
+            return None
+        elif isinstance(exp_set, str):
+            if exp_set.isdigit():
+                # convert string containing an integer
+                exp_set = int(exp_set)
+            elif exp_set.lower() != 'sum':
+                print(f"ERROR: invalid raster string of '{exp_set}'. Please"
+                     +f" input either 'sum' or an integer between 0 and"
+                     +f" {nexp_per_pos-1}.", file=sys.stderr)
+                return None
+
+        if isinstance(exp_set, int):
+            if (exp_set >= nexp_per_pos) or (exp_set < -1*nexp_per_pos):
+                # Check for out-of-bound indices
+                oob_exp_set = exp_set
+                if exp_set < 0:
+                    exp_set = 0
+                else:
+                    exp_set = nexp_per_pos-1
+                print(f"WARNING: exp_set {oob_exp_set} is out of bounds."
+                     +f" Loading the nearest valid exp_set ({exp_set}) instead.", 
+                     file=sys.stderr)
+            elif exp_set < 0:
+                # convert negative indices to positive
+                exp_set = nexp_per_pos - exp_set
+    
+        # Extract the correct subarrays
+        if str(exp_set).lower() == 'sum':
+            print(f'Summing all {nexp_per_pos} sets exposures per slit position')
+            lv_1_counts = np.nansum(lv_1_counts, axis=0)
+            meta['date_obs'] = meta['date_obs'][0, :]
+            meta['duration'] = np.nansum(meta['duration'], axis=0)
+            meta['wave_corr'] = np.nanmean(meta['wave_corr'], axis=0)
+            meta['wave_corr_t'] = np.nanmean(meta['wave_corr_t'], axis=0)
+            meta['pointing']['solar_x'] = meta['pointing']['solar_x'][0, :]
+        else:
+            print(f'Loading exposure set {exp_set} ({exp_set+1} of {nexp_per_pos})')
+            lv_1_counts = lv_1_counts[exp_set,:,:,:]
+            meta['date_obs'] = meta['date_obs'][exp_set, :]
+            meta['duration'] = meta['duration'][exp_set, :]
+            meta['wave_corr'] = meta['wave_corr'][exp_set, :, :]
+            meta['wave_corr_t'] = meta['wave_corr_t'][exp_set, :]
+            meta['pointing']['solar_x'] = meta['pointing']['solar_x'][exp_set, :]
+    elif len(lv_1_counts.shape) > 4:
+        print(f"ERROR: cannot read input file with data array shape of"
+             +f" {lv_1_counts.shape}.", file=sys.stderr)
+        return None
+
+    ############################################################################
+    ### Determine observation type
+    ############################################################################
+    if index['nraster'] == 1:
+        obs_type = 'sit-and-stare'
+    elif index['nexp_prp'] >= 2:
+        obs_type = 'multi_scan'
+    else:
+        obs_type = 'scan'
+
     ############################################################################
     ### Apply pointing corrections and create output EISCube
     ############################################################################
@@ -226,10 +316,6 @@ def read_cube(filename=None, window=0, apply_radcal=True, radcal=None,
 
     ### (3) Apply wave correction and get median base wavelength value and delta
     counts_shape = lv_1_counts.shape
-    if len(counts_shape) > 3:
-        print('ERROR: EIS observations with multiple rasters in a single file'
-             +' are not currently supported.', file=sys.stderr)
-        return None
     ny_pxls = counts_shape[0] # num pixels along the slit (y-axis)
     nx_steps = counts_shape[1] # num raster steps (x-axis)
     n_wave = counts_shape[2] # num wavelength values (i.e. EIS window width)
@@ -257,10 +343,20 @@ def read_cube(filename=None, window=0, apply_radcal=True, radcal=None,
     ### (6) Fetch the observer location in heliographic coords
     hg_coords = coords.get_body_heliographic_stonyhurst('earth', time=date_obs)
 
-    ### (7) Create a new header dict updated values
+    ### (7) Calculate average exposure time and cadence
+    avg_exptime = np.mean(meta['duration'])
+    if index['nexp'] > 1:
+        diff_date_obs = np.diff(meta['date_obs'].astype('datetime64'))
+        diff_date_obs = diff_date_obs / np.timedelta64(1, 's')
+        avg_cad = np.mean(np.abs(diff_date_obs))
+    else:
+        avg_cad = avg_exptime # default for single exposure obs
+
+    ### (8) Create a new header dict updated values
     # Note: the order of axes here is the same as the original fits index
     output_hdr = dict()
 
+    # Time information
     output_hdr['naxis'] = 3
     output_hdr['date_obs'] = date_obs
     output_hdr['date_beg'] = date_obs
@@ -268,19 +364,39 @@ def read_cube(filename=None, window=0, apply_radcal=True, radcal=None,
     output_hdr['date_end'] = date_end
     output_hdr['timesys'] = 'UTC'
 
+    # Observation details
     output_hdr['telescop'] = 'Hinode'
     output_hdr['instrume'] = 'EIS'
-    output_hdr['stud_acr'] = index['stud_acr']
-    output_hdr['obstitle'] = index['obstitle']
-    output_hdr['target'] = index['target']
-    output_hdr['sci_obj'] = index['sci_obj']
     output_hdr['line_id'] = wininfo['line_id'][meta['iwin']]
     output_hdr['measrmnt'] = 'intensity'
     output_hdr['bunit'] = 'unknown' # units of primary observable
     output_hdr['slit_id'] = index['slit_id']
     output_hdr['slit_ind'] = index['slit_ind']
-    output_hdr['nraster'] = index['nraster']
+    output_hdr['obs_type'] = obs_type
+    output_hdr['nraster'] = index['nraster'] # 1 for sit-and-stare
+    output_hdr['nexp'] = index['nexp'] # total num exposures
+    output_hdr['nexp_prp'] = index['nexp_prp'] # num exposures per slit position
+    output_hdr['tr_mode'] = index['tr_mode'] # tracking mode. "FIX" for no tracking
+    output_hdr['saa'] = index['saa'] # IN / OUT South Atlantic Anomaly (SAA)
+    output_hdr['hlz'] = index['hlz'] # IN / OUT High-Latitude Zone (HLZ)
+    output_hdr['exptime'] = avg_exptime
+    output_hdr['cadence'] = avg_cad
+    output_hdr['timeunit'] = 's'
 
+    # IDs and Study information
+    output_hdr['tl_id'] = index['tl_id']
+    output_hdr['jop_id'] = index['jop_id']
+    output_hdr['study_id'] = index['study_id']
+    output_hdr['stud_acr'] = index['stud_acr']
+    output_hdr['rast_id'] = index['rast_id']
+    output_hdr['rast_acr'] = index['rast_acr']
+    output_hdr['obstitle'] = index['obstitle'][:-1].strip()
+    output_hdr['obs_dec'] = index['obs_dec'][:-1].strip()
+    output_hdr['target'] = index['target']
+    output_hdr['sci_obj'] = index['sci_obj'][:-1].strip()
+    output_hdr['noaa_num'] = index['noaa_num']
+
+    # Coordinate information
     output_hdr['naxis1'] = nx_steps
     output_hdr['cname1'] = 'Solar-X'
     output_hdr['crval1'] = x1
@@ -326,9 +442,7 @@ def read_cube(filename=None, window=0, apply_radcal=True, radcal=None,
         # NB: For some reason, the order of axes in the WCS is reversed relative
         #     to the data array inside an NDCube. We should take care to fully
         #     document this for our users.
-        clean_wcs = astropy.wcs.WCS(output_hdr, fix=True)
-        clean_wcs = clean_wcs.swapaxes(0,1) # swap x with y
-        clean_wcs = clean_wcs.swapaxes(0,2) # now swap y with wavelength
+        clean_wcs = _make_wcs_from_cube_index(output_hdr)
 
         # Add a user-supplied constant value to the count array
         if count_offset is not None:
@@ -356,6 +470,7 @@ def read_cube(filename=None, window=0, apply_radcal=True, radcal=None,
             cube_errs = lv_1_count_errs
             data_units = 'photon'
 
+        cube_errs = StdDevUncertainty(cube_errs)
         meta['mod_index']['bunit'] = data_units
         output_cube = EISCube(cube_data, wcs=clean_wcs, uncertainty=cube_errs,
                               wavelength=corrected_wave, radcal=radcal_array,
